# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import json
from enum import Enum

from typing import Callable, Any, Optional

from pypz.core.commons.loggers import ContextLogger, DefaultContextLogger
from pypz.core.commons.utils import current_time_millis


NameSeparator = "@"
"""
Used to construct channel unique name
"""


class ChannelFilter:
    """
    This interface contains the methods that shall implement the checking of different
    statuses. This interface can be used to restrict access to ChannelStatusMonitor
    object.
    """

    def is_channel_not_initialised(self) -> bool:
        pass

    def is_channel_opened(self) -> bool:
        pass

    def is_channel_started(self) -> bool:
        pass

    def is_channel_stopped(self) -> bool:
        pass

    def is_channel_closed(self) -> bool:
        pass

    def is_channel_error(self) -> bool:
        pass

    def is_channel_healthy(self) -> bool:
        pass

    def get_channel_unique_name(self) -> str:
        pass

    def get_channel_context_name(self) -> str:
        pass

    def get_channel_group_name(self) -> str:
        pass


class ChannelStatus(Enum):
    Opened = "Opened"
    Started = "Started"
    Stopped = "Stopped"
    Closed = "Closed"
    HealthCheck = "HealthCheck"
    Error = "Error"
    WaitingForSchema = "WaitingForSchema"
    Acknowledged = "Acknowledged"


class ChannelStatusMessage:
    """
    This class is a Data Transfer Object (DTO) for the status messages sent between the channels.

    :param channel_name: name of the channel
    :param channel_context_name: name of the context of the channel, which is the full name of the plugin
    :param channel_group_name: name of the group of the channel, which is the group name of the plugin
    :param channel_group_index: index of the operator in the group, which is inherited by the plugin
    :param status: status
    :param payload: optional payload
    :param timestamp: timestamp of the message at creation
    """

    def __init__(self,
                 channel_name: str,
                 channel_context_name: str,
                 channel_group_name: str = None,
                 channel_group_index: int = 0,
                 status: ChannelStatus | str = None,
                 payload: Any = None,
                 timestamp: int = None):
        self.channel_name: str = channel_name
        self.channel_context_name: str = channel_context_name
        self.channel_group_name: str = channel_group_name
        self.channel_group_index: int = channel_group_index
        self.status: ChannelStatus = status if isinstance(status, ChannelStatus) else ChannelStatus[status]
        self.payload: Any = payload
        self.timestamp: int = timestamp if timestamp is not None else current_time_millis()

    def get_channel_unique_name(self):
        return f"{self.channel_name}@{self.channel_context_name}"

    @staticmethod
    def create_from_json(source):
        if isinstance(source, str):
            return ChannelStatusMessage(**json.loads(source))

        if isinstance(source, dict):
            return ChannelStatusMessage(**source)

        raise TypeError("Invalid data type provided for json conversion: " + str(type(source)))

    def __str__(self):
        return json.dumps(self, default=lambda k: list(k) if isinstance(k, set) else k.__dict__ if not isinstance(k, Enum) else k.value)


class ChannelStatusMonitor(ChannelFilter):
    """
    This class has the responsibility to monitor and track state of the connected channel RW entities.
    If the counterpart sends a status message, this class has to parse and interpret it. In addition,
    it provides an implementation for the ``ChannelFilter`` interface to enable certain queries.

    :param channel_name: name of the channel
    :param channel_context_name: name of the context of the channel, which is the full name of the plugin
    :param channel_group_name: name of the group of the channel, which is the group name of the plugin
    :param logger: the context logger of the plugin, which is inherited from the operator
    """

    ValidPatienceTimeInMs = 120000
    """
    This is the patience time for not receiving any update from the ChannelReader. Once this time
    is up, the ChannelReader will be invalidated
    """

    def __init__(self,
                 channel_name: str,
                 channel_context_name: str,
                 channel_group_name: str = None,
                 logger: ContextLogger = None):
        self.channel_name: str = channel_name

        self.channel_context_name: str = channel_context_name

        self.channel_group_name: str = channel_group_name

        self.status_update_map: dict[ChannelStatus, int] = dict()

        self.current_status: Optional[ChannelStatus] = None

        self.last_status_change_timestamp_ms: int = 0

        self.last_update_sent_timestamp_ms: int = 0

        self.monitor_start_timestamp: int = current_time_millis()

        self.status_update_callback: Optional[Callable[[ChannelStatusMessage], None]] = None

        self.logger: ContextLogger = \
            ContextLogger(logger, *logger.get_context_stack(), channel_context_name) \
            if logger is not None else ContextLogger(DefaultContextLogger(self.get_channel_unique_name()))

    def get_channel_unique_name(self) -> str:
        return f"{self.channel_name}{NameSeparator}{self.channel_context_name}"

    def get_channel_context_name(self) -> str:
        return self.channel_context_name

    def get_channel_group_name(self) -> str:
        return self.channel_group_name

    def is_channel_healthy(self) -> bool:
        return ChannelStatusMonitor.ValidPatienceTimeInMs > (current_time_millis() - self.monitor_start_timestamp) \
            if 0 == self.last_update_sent_timestamp_ms else \
            ChannelStatusMonitor.ValidPatienceTimeInMs > (current_time_millis() - self.last_update_sent_timestamp_ms)

    def is_channel_not_initialised(self):
        return self.current_status is None

    def is_channel_opened(self):
        return self.current_status == ChannelStatus.Opened

    def is_channel_closed(self):
        return self.current_status == ChannelStatus.Closed

    def is_channel_started(self):
        return self.current_status == ChannelStatus.Started

    def is_channel_stopped(self):
        return self.current_status == ChannelStatus.Stopped

    def is_channel_error(self):
        return self.current_status == ChannelStatus.Error

    def on_status_update(self, statusUpdateCallback: Callable):
        self.status_update_callback = statusUpdateCallback

    def update(self, status_message: ChannelStatusMessage):
        """
        This method implements the logic to handle status message updates.
        Note that to overcome messaging issues e.g., missing status update,
        we don't introduce a state machine. It means that most of the states
        can overwrite the actual state with one exception. If "error" status
        has been sent, it can only be overwritten by "open" or "start". This
        allows us to retain error information until the counterpart has been
        restarted.

        :param status_message: the status message received from the other ChannelRW entity
        """

        if status_message.get_channel_unique_name() != self.get_channel_unique_name():
            raise AttributeError(f"Invalid status message, "
                                 f"mismatching channel name: {status_message.get_channel_unique_name()}")

        # int has been merged with long https://www.python.org/dev/peps/pep-0237/
        status_update_time = int(float(status_message.timestamp))

        if status_update_time > self.last_update_sent_timestamp_ms:
            self.last_update_sent_timestamp_ms = status_update_time
        elif status_update_time < self.last_update_sent_timestamp_ms:
            self.logger.warn(f"WARNING - the current status message has lower timestamp "
                             f"({status_message.timestamp}) than the last received "
                             f"({self.last_update_sent_timestamp_ms}). This may be caused by network latency, "
                             f"but if you repeatedly see this warning, then you should investigate the causes. "
                             f"({status_message})")

        new_status = self.current_status

        if (ChannelStatus.Opened == status_message.status) or \
           (ChannelStatus.Closed == status_message.status) or \
           (ChannelStatus.Started == status_message.status) or \
           (ChannelStatus.Stopped == status_message.status) or \
           (ChannelStatus.Error == status_message.status):
            if status_update_time >= self.last_status_change_timestamp_ms:
                if ChannelStatus.Error != self.current_status:
                    new_status = status_message.status
                elif (ChannelStatus.Opened == status_message.status) or \
                     (ChannelStatus.Started == status_message.status):
                    new_status = status_message.status

        if (new_status is None) or (new_status != self.current_status):
            self.logger.debug(f"{self.current_status} -> {new_status} @ {status_update_time}")
            self.current_status = new_status
            self.last_status_change_timestamp_ms = status_update_time

        if status_message.payload is not None:
            self.logger.debug(f"Message payload: {status_message.payload}")

        if self.status_update_callback is not None:
            self.status_update_callback(status_message)

        self.status_update_map[status_message.status] = status_update_time





# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================


class BaseSignal:
    def __init__(self):
        pass


class SignalError(BaseSignal):
    def __init__(self, exc: Exception = None):
        super().__init__()
        self.__exc = exc

    def get_exception(self):
        return self.__exc


class SignalKill(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalNoOp(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationInit(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationStart(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalOperationStop(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalResourcesCreation(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalResourcesDeletion(BaseSignal):

    def __init__(self):
        super().__init__()


class SignalServicesStart(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalServicesStop(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalShutdown(BaseSignal):
    def __init__(self):
        super().__init__()


class SignalTerminate(BaseSignal):
    def __init__(self):
        super().__init__()


# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import Optional, Type, TypeVar

NestedInstanceDTOType = TypeVar("NestedInstanceDTOType", bound='InstanceDTO')


class SpecDTO:
    """
    This class represents the base Data Transfer Object for an instance spec.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the specs
    :param location: the location of the spec to retrieve it from
    :param expectedParameters: the expected parameters by the spec
    :param types: the implemented types of the spec
    :param nestedInstanceType: the expected type of the nested instances
    :param nestedInstances: the list of the actual nested instances
    :param nested_instance_dto_type: helper to specify the nested DTO type
    """

    def __init__(self,
                 name: str = None,
                 location: str = None,
                 expectedParameters: dict = None,
                 types: list = None,
                 nestedInstanceType: str = None,
                 nestedInstances: list[NestedInstanceDTOType] = None,
                 nested_instance_dto_type: Optional[Type[NestedInstanceDTOType]] = 'default'):
        self.name: str = name
        self.location: str = location
        self.expectedParameters: dict = expectedParameters
        self.types: list = types
        self.nestedInstanceType: str = nestedInstanceType
        self.nestedInstances: Optional[set[NestedInstanceDTOType]] = None

        if nested_instance_dto_type == 'default':
            nested_instance_dto_type = InstanceDTO

        if (nestedInstances is not None) and (nested_instance_dto_type is not None):
            self.nestedInstances: list[NestedInstanceDTOType] = list()
            if isinstance(nestedInstances, (set, list)):
                for nestedInstance in nestedInstances:
                    if isinstance(nestedInstance, InstanceDTO):
                        self.nestedInstances.append(nestedInstance)
                    elif isinstance(nestedInstance, dict):
                        self.nestedInstances.append(nested_instance_dto_type(**nestedInstance))
                    else:
                        raise TypeError(f"Invalid nestedInstance type: {type(nestedInstance)}")
            else:
                raise TypeError(f"Invalid nestedInstances type: {type(nestedInstances)}")

    def __eq__(self, other):
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.name == other.name) and \
            (self.location == other.location) and \
            (self.expectedParameters == other.expectedParameters) and \
            (self.types == other.types) and \
            (self.nestedInstanceType == other.nestedInstanceType) and \
            (self.nestedInstances == other.nestedInstances)

    def __hash__(self):
        return hash(self.name)


class InstanceDTO:
    """
    This class represents the base Data Transfer Object for an instance.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the instance
    :param parameters: parameters of the instance
    :param dependsOn: list of instances that this instance depends on
    """

    def __init__(self, name: str = None, parameters: dict = None, dependsOn: list[str] = None,
                 spec: SpecDTO = None):
        self.name: str = name
        self.parameters: dict = parameters
        self.dependsOn: list[str] = dependsOn

        if (spec is None) or isinstance(spec, SpecDTO):
            self.spec = spec
        elif isinstance(spec, dict):
            self.spec = SpecDTO(**spec)
        else:
            raise TypeError(f"Invalid instance spec type: {type(spec)}")

    def __eq__(self, other):
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.name == other.name) and \
            (self.parameters == other.parameters) and \
            (self.dependsOn == other.dependsOn) and \
            (self.spec == other.spec)

    def __hash__(self):
        return hash((self.name, self.spec.__hash__()))


class PluginSpecDTO(SpecDTO):
    """
    This class represents the Data Transfer Object for a plugin spec.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the specs
    :param location: the location of the spec to retrieve it from
    :param expectedParameters: the expected parameters by the spec
    :param types: the implemented types of the spec
    :param nestedInstanceType: the expected type of the nested instances
    :param nestedInstances: the list of the actual nested instances
    """

    def __init__(self,
                 name: str = None,
                 location: str = None,
                 expectedParameters: dict = None,
                 types: list = None,
                 nestedInstanceType: str = None,
                 nestedInstances: list = None):
        super().__init__(name, location, expectedParameters, types, nestedInstanceType, None, None)


class PluginInstanceDTO(InstanceDTO):
    """
    This class represents the base Data Transfer Object for a plugin instance.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the instance
    :param parameters: parameters of the instance
    :param dependsOn: list of instances that this instance depends on
    """

    def __init__(self,
                 name: str = None,
                 parameters: dict = None,
                 dependsOn: list[str] = None,
                 spec: PluginSpecDTO = None):
        super().__init__(name, parameters, dependsOn, None)

        if (spec is None) or isinstance(spec, PluginSpecDTO):
            self.spec = spec
        elif isinstance(spec, dict):
            self.spec = PluginSpecDTO(**spec)
        else:
            raise TypeError(f"Invalid instance spec type: {type(spec)}")


class OperatorSpecDTO(SpecDTO):
    """
    This class represents the Data Transfer Object for an operator spec.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the specs
    :param location: the location of the spec to retrieve it from
    :param expectedParameters: the expected parameters by the spec
    :param types: the implemented types of the spec
    :param nestedInstanceType: the expected type of the nested instances
    :param nestedInstances: the list of the actual nested instances
    """

    def __init__(self,
                 name: str = None,
                 location: str = None,
                 expectedParameters: dict = None,
                 types: list = None,
                 nestedInstanceType: str = None,
                 nestedInstances: list[PluginInstanceDTO] = None):
        super().__init__(name, location, expectedParameters, types, nestedInstanceType, nestedInstances, PluginInstanceDTO)


class OperatorConnectionSource:
    """
    This class represents the base Data Transfer Object for the source of
    a connection between operator instances.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param instanceName: name of the source instance
    :param outputPortName: name of the output port in the source instance
    """

    def __init__(self,
                 instanceName: str = None,
                 outputPortName: str = None):
        self.instanceName: str = instanceName
        self.outputPortName: str = outputPortName

    def __eq__(self, other):
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.instanceName == other.instanceName) and \
            (self.outputPortName == other.outputPortName)

    def __hash__(self):
        return hash((self.instanceName, self.outputPortName))


class OperatorConnection:
    """
    This class represents the base Data Transfer Object for a connection between operator instances.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param inputPortName: name of the input port of the connection
    :param source: source DTO of the connection
    """

    def __init__(self,
                 inputPortName: str = None,
                 source: dict | OperatorConnectionSource = None):
        self.inputPortName: str = inputPortName
        self.source: OperatorConnectionSource = source \
            if isinstance(source, OperatorConnectionSource) \
            else OperatorConnectionSource(**source)

    def __eq__(self, other):
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.inputPortName == other.inputPortName) and \
            (self.source == other.source)

    def __hash__(self):
        return hash((self.inputPortName, self.source.__hash__()))


class OperatorInstanceDTO(InstanceDTO):
    """
    This class represents the base Data Transfer Object for an operator instance.
    Operator instances have additional instance
    information that shall be modelled via the DTO e.g., since Operator
    instances can be connected through port plugins, those connections shall
    be modelled as well.

    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the instance
    :param parameters: parameters of the instance
    :param dependsOn: list of instances that this instance depends on
    """

    def __init__(self,
                 name: str = None,
                 parameters: dict = None,
                 dependsOn: list[str] = None,
                 spec: OperatorSpecDTO = None,
                 connections: list[OperatorConnection] = None):
        super().__init__(name, parameters, dependsOn, None)

        if connections is not None:
            self.connections: Optional[list[OperatorConnection]] = list()
            if isinstance(connections, (set, list)):
                for connection in connections:
                    if isinstance(connection, OperatorConnection):
                        self.connections.append(connection)
                    elif isinstance(connection, dict):
                        self.connections.append(OperatorConnection(**connection))
                    else:
                        raise TypeError(f"Invalid nestedInstance type: {type(connection)}")
            else:
                raise TypeError(f"Invalid nestedInstances type: {type(connections)}")

        else:
            self.connections: Optional[set[OperatorConnection]] = None

        if (spec is None) or isinstance(spec, OperatorSpecDTO):
            self.spec = spec
        elif isinstance(spec, dict):
            self.spec = OperatorSpecDTO(**spec)
        else:
            raise TypeError(f"Invalid instance spec type: {type(spec)}")

    def __eq__(self, other):
        if self is other:
            return True

        return super().__eq__(other) and \
            (self.connections == other.connections)

    def __hash__(self):
        return super().__hash__()


class PipelineSpecDTO(SpecDTO):
    """
    This class represents the Data Transfer Object for a pipeline spec.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the specs
    :param location: the location of the spec to retrieve it from
    :param expectedParameters: the expected parameters by the spec
    :param types: the implemented types of the spec
    :param nestedInstanceType: the expected type of the nested instances
    :param nestedInstances: the list of the actual nested instances
    """

    def __init__(self,
                 name: str = None,
                 location: str = None,
                 expectedParameters: dict = None,
                 types: list = None,
                 nestedInstanceType: str = None, nestedInstances: list[OperatorInstanceDTO] = None):
        super().__init__(name, location, expectedParameters, types, nestedInstanceType, nestedInstances, OperatorInstanceDTO)


class PipelineInstanceDTO(InstanceDTO):
    """
    This class represents the base Data Transfer Object for a pipeline instance.
    Note that using the coding language as analogy, the instance spec is like
    a class, where the instance itself is like the object created from the class.
    If we want to transfer the instance spec remotely, we need to convert it
    into this DTO representation. This class acts as base for every other specs.

    .. note::
       Notice that the names of the attributes and ctor arguments will be used for
       serialization, hence those are not following python's format guidelines.

    :param name: name of the instance
    :param parameters: parameters of the instance
    :param dependsOn: list of instances that this instance depends on
    """

    def __init__(self,
                 name: str = None,
                 parameters: dict = None,
                 dependsOn: list[str] = None,
                 spec: PipelineSpecDTO = None):
        super().__init__(name, parameters, dependsOn, None)

        if (spec is None) or isinstance(spec, PipelineSpecDTO):
            self.spec = spec
        elif isinstance(spec, dict):
            self.spec = PipelineSpecDTO(**spec)
        else:
            raise TypeError(f"Invalid instance spec type: {type(spec)}")



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import logging
import threading
import time
import traceback
from typing import Callable, Any, TYPE_CHECKING
from abc import ABC, abstractmethod
import concurrent.futures
import json

from pypz.core.channels.status import ChannelStatusMonitor, ChannelFilter, ChannelStatusMessage, ChannelStatus, \
    NameSeparator
from pypz.core.commons.loggers import ContextLogger
from pypz.core.commons.utils import ensure_type, current_time_millis, SynchronizedReference

if TYPE_CHECKING:
    from pypz.core.specs.plugin import PortPlugin


class ChannelMetric:
    def __init__(self, elapsedTimeSinceLastIO, recordCountInLastIO):
        self.elapsedTimeSinceLastIO = elapsedTimeSinceLastIO
        self.recordCountInLastIO = recordCountInLastIO


class ChannelBase(ABC):
    """
    The abstract base for the :class:`ChannelReader <pypz.core.channels.io.ChannelReader>` and the
    :class:`ChannelWriter <pypz.core.channels.io.ChannelWriter>` classes. It contains the common
    logic to be used in both classes' implementations.

    :param channel_name: name of the channel
    :param context: the :class:`PortPlugin <pypz.core.specs.plugin.PortPlugin>`, which operates this channel
    :param executor: an external ThreadPoolExecutor, if not provided, on will be created internally
    """

    # ======================= Static fields =======================

    DefaultStatusThreadIntervalInMs = 2000
    """
    This value defines in how many ms the status messages (health check) will be sent to the
    counterpart channel.
    """

    ControlLoopExceptionTimeoutInMs = 60000
    """
    This value defines the timeout for having unhandled exception in the control loop.
    After expiration, the control loop will be terminated.    
    """

    ParamKeyLogLevel = "logLevel"
    ParamKeyMetricsEnabled = "metricsEnabled"

    # ======================= Ctor =======================

    def __init__(self, channel_name: str,
                 context: 'PortPlugin',
                 executor: concurrent.futures.ThreadPoolExecutor = None,
                 **kwargs):

        if channel_name is None:
            raise ValueError("Channel name must be provided")

        if context is None:
            raise ValueError("Context instance must be provided")

        self.__silent_mode: bool = kwargs["silent_mode"] if "silent_mode" in kwargs else False
        """
        If this flag is set, then this channel will not send status messages. One use-case is,
        if a channelRW is created to sniff the status of channels.
        """

        self._channel_name: str = channel_name
        """
        This member stores the name of the channel, which normally reflects to the resource's name.
        """

        self._context: PortPlugin = context
        """
        The context of this channel, which shall be an object of PortPlugin type
        """
        
        self._unique_name: str = f"{self._channel_name}{NameSeparator}{self._context.get_full_name()}"
        """
        This name identifies the channel in its context. It is necessary, since channel name on
        its own is not unique, can be reused by different owners.
        """

        self._location: str | None = None
        """
        This member stores the location string of the channel. The location depends on the technology. It can refer
        to either a remote or a local URL as well.
        """

        self._channel_opened: bool = False
        """
        This member is a flag to signalize, whether a channel is open and ready to use.
        """

        self._channel_started: bool = False
        """
        Status flag to signalize, whether the channel has been started.
        """

        self._channel_stopped: bool = False
        """
        Status flag to signalize, whether the channel has been stopped. Note that this additional flag is
        necessary, since having the _channel_started on false does not necessarily mean that the
        output was ever started and is now finished.        
        """

        self._resources_created: bool = False
        """
        Status flag to signalize that the channel's resources are created
        """

        self._resources_deleted: bool = False
        """
        Status flag to signalize that the channel's resources are deleted
        """

        self._executor_started: bool = False
        """
        Flag to signalize, whether the executor thread started
        """

        self._stopping_executor: bool = False
        """
        Flag to signalize the termination attempt of the executor thread
        """

        self._executor_stopped: SynchronizedReference[bool] = SynchronizedReference(False)
        """
        Flag to signalize that the executor thread has been terminated. This flag is necessary along wi
        _stopping_executor to synchronize the termination. Otherwise it can happen that the 
        thread did not terminate yet, but the channel has already been closed resulting in exceptions.
        """

        self._configuration: dict = dict()
        """
        This member stores the configuration string of the channel. This is a kind of serialized storage, since
        the configuration can be an arbitrary type with arbitrary schema as long as the interpretation is knonw.
        """

        self._status_map: dict[str, ChannelStatusMonitor] = dict()
        """
        This map stores the health statuses of all connected channels
        """

        self._executor: concurrent.futures.ThreadPoolExecutor = executor
        """
        This executor is used to execute the background thread, which will invoke the sendHealthStatus method
        """

        self._control_loop_exception_timer: int = 0
        """
        Control loop is, where the control information are sent to the counter channel. Should
        there an exception occur, we need to make sure that we give some time to recover, before
        terminating the channel. This timer defines that grace period.
        """

        self._health_check_payload: dict[str, Any] = dict()
        """
        It is possible to send additional information along with health check events. This
        member serves as the storage of this additional payload.
        """

        self._metrics_enabled: bool = False
        """
        Enables metric calculation i.e., to calculate additional i/o related information
        """

        self._channel_state_update_lock = threading.Lock()
        """
        This lock is used to synchronize the logic that updates channel states
        """

        self._on_status_message_received_callbacks: set[Callable[[list[ChannelStatusMessage]], None]] = set()
        """
        Stores the callbacks, which will be executed, if status messages received
        """

        self._logger: ContextLogger = ContextLogger(self._context.get_logger(),
                                                    self._context.get_full_name(),
                                                    self._channel_name)
        """
        Channel logger
        """

        self._log_level: str = "DEBUG"
        """
        Default log level. This is ok to set here, since if Operator's context logger is present,
        its log level cannot be overriden, but if not, the DefaultContextLogger's can be.
        """

        self._logger.set_log_level(logging.DEBUG)

    # ======================= Abstract methods =======================

    @abstractmethod
    def _create_resources(self) -> bool:
        """
        This method shall implement the logic of creating resources of the channel.
        IMPORTANT NOTE - this method shall be invoked before the open_channel() to make sure that the resources
        are existing beforehand.

        :return: True, if done, False if it is still in progress
        """
        pass

    @abstractmethod
    def _delete_resources(self) -> bool:
        """
        This method shall implement the logic of deleting resources of the channel.
        IMPORTANT NOTE - this method shall be invoked after the close_channel()

        :return: True, if done, False if it is still in progress
        """
        pass

    @abstractmethod
    def _open_channel(self) -> bool:
        """
        This method shall implement the logic to open a channel. The meaning of 'open' however is to be defined
        by the actual implementation. One developer can define it like an opened connection, other as created file etc.

        :return: True, if done, False if it is still in progress
        """
        pass

    @abstractmethod
    def _close_channel(self) -> bool:
        """
        This method shall implement the logic to close a channel. Normally closing a channel is the last step so
        clean up of used resource shall happen here as well.

        :return: True, if done, False if it is still in progress
        """
        pass

    @abstractmethod
    def _configure_channel(self, channel_configuration: dict) -> None:
        """
        This method shall implement the logic to interpret the provided configuration.

        :param channel_configuration: config string
        """
        pass

    @abstractmethod
    def _send_status_message(self, message: str) -> None:
        """
        This method shall implement the logic that publishes the channel's state to the counterpart channel. The
        state string is provided by the channel itself. Note that there is defined schema, how and what will be
        provided by the channel as string, however you can append your own custom information, you only need to
        append as string separated by StateMessageSeparatorChar.

        :param message: message that shall be sent
        """
        pass

    @abstractmethod
    def _retrieve_status_messages(self) -> list:
        """
        This method shall implement the logic that retrieves the status messages published by the counterpart
        channel. Notice that in case of ChannelWriter, there can be multiple InputChannels sending messages and
        in case of ChannelReader there can be multiple messages sent by the ChannelWriter, therefore this method
        shall return a list of messages. Note as well that in case you are using callbacks for your technology,
        you can directly use the method onStatusMessageReceived. In this case simply return null from this method
        to ensure that it will be ignored.

        :return: list of retrieved status messages or null if method is not used
        """
        pass

    # ======================= Getter/setter/invoker methods =======================

    def get_logger(self):
        return self._logger

    def get_channel_name(self):
        return self._channel_name

    def get_unique_name(self):
        return self._unique_name

    def get_location(self):
        return self._location

    def get_context(self) -> 'PortPlugin':
        return self._context

    def get_configuration(self) -> dict:
        return self._configuration

    def is_channel_open(self):
        return self._channel_opened

    def is_channel_started(self):
        return self._channel_started

    def is_channel_stopped(self):
        return self._channel_stopped

    def is_resource_created(self):
        return self._resources_created

    def is_resource_deleted(self):
        return self._resources_deleted

    def set_location(self, channel_location: str):
        self._location = channel_location

    def retrieve_all_connected_channel_count(self) -> int:
        """
        Returns the number of tracked input channels. This can be used to synchronize some activities like starting
        the output.

        :return: number of tracked input channels
        """

        return len(self._status_map)

    def retrieve_connected_channel_unique_names(self, filter_function: Callable[[ChannelFilter], bool] = None) -> set:
        """
        This method returns a set of connected channel names given the evaluation
        criteria passed as argument.

        :param filter_function: check_function check the ChannelFilter for more details (nullable)
        :return: Set of channel names that passes the given evaluation
        """

        connected_channel_names = set()

        for key, val in self._status_map.items():
            if filter_function is not None:
                if filter_function(val):
                    connected_channel_names.add(key)
            else:
                connected_channel_names.add(key)

        return connected_channel_names

    def retrieve_healthy_connected_channel_count(self) -> int:
        """
        Returns the number of healthy tracked input channels.
        This can be used to synchronize some activities like starting the output.

        :return: number of healthy tracked input channels
        """

        healthy_connected_channel_count = 0

        for statusMonitor in list(self._status_map.values()):
            if statusMonitor.is_channel_healthy():
                healthy_connected_channel_count += 1

        return healthy_connected_channel_count

    def is_any_connected_channel_healthy_and_not_stopped(self) -> bool:
        """
        This method realises the question "Is there any connected channel that is healthy, but not
        stopped?". This can be seen as opposite of "Is all connected channels stopped or unhealthy?"
        with the strong difference that the latter does not give a proper answer if there is no
        connected channels at all.

        :return: True if there is any channel healthy, but not stopped, False if not or no connected channel
        """

        for status_monitor in list(self._status_map.values()):
            if status_monitor.is_channel_healthy() and not status_monitor.is_channel_stopped():
                return True
        return False

    def is_any_connected_channel_healthy_and_not_closed(self) -> bool:
        """
        This method realises the question "Is there any connected channel that is healthy, but not
        closed?". This can be seen as opposite of "Is all connected channels closed or unhealthy?"
        with the strong difference that the latter does not give a proper answer if there is no
        connected channels at all.

        :return: True if there is any channel healthy, but not closed, False if not or no connected channel
        """

        for status_monitor in list(self._status_map.values()):
            if status_monitor.is_channel_healthy() and not status_monitor.is_channel_closed():
                return True
        return False

    def is_any_connected_channel_healthy_and_not_stopped_and_not_closed(self) -> bool:
        """
        This method realises the question "Is there any connected channel that is healthy, but not
        stopped and not closed?". This can be seen as opposite of "Is all connected channels closed or
        stopped or unhealthy?" with the strong difference that the latter does not give a proper
        answer if there is no connected channels at all.

        :return: True if there is any channel healthy, but not stopped and not closed,
                 False if not or no connected channel
        """

        for status_monitor in list(self._status_map.values()):
            if status_monitor.is_channel_healthy() and \
                    not status_monitor.is_channel_stopped() and \
                    not status_monitor.is_channel_closed():
                return True
        return False

    def is_any_connected_channels_unhealthy(self) -> bool:
        """
        This method realises the question "Is there any connected channels unhealthy?". This can be seen
        as opposite of "Is all connected channels healthy?" with the strong difference that the latter
        does not give a proper answer if there is no connected channels at all.

        :return: True if there is any channel unhealthy, False if not or no connected channels
        """

        for status_monitor in list(self._status_map.values()):
            if not status_monitor.is_channel_healthy():
                return True
        return False

    def is_any_connected_channels_healthy(self) -> bool:
        """
        This method realises the question "Is there any channel healthy?". This can be seen as opposite
        of "Is all channels unhealthy?" with the strong difference that the latter does not give a proper
        answer if there is no connected channels at all.

        :return: True if there is any channel healthy, False if not or no connected channel
        """
        for status_monitor in list(self._status_map.values()):
            if status_monitor.is_channel_healthy():
                return True
        return False

    def start_channel(self, send_status_message: bool = True) -> None:
        """
        This method sets the corresponding flags and sends the corresponding status message to signalize that
        the channel has been started.

        :param send_status_message: if True the status message sending will be called
        """

        if send_status_message:
            self.invoke_sync_send_status_message(ChannelStatus.Started)

        self._channel_started = True
        self._channel_stopped = False

    def stop_channel(self, send_status_message: bool = True) -> None:
        """
        This method sets the corresponding flags and sends the corresponding status message to signalize that
        the channel has been stopped.

        :param send_status_message: if True the status message sending will be called
        """
        self._channel_started = False
        self._channel_stopped = True

        if send_status_message:
            self.invoke_sync_send_status_message(ChannelStatus.Stopped)

    def can_close(self) -> bool:
        """
        This method can implement the logic to determine, if the channel can
        be closed or not.

        :return: True if the channel can be closed, False otherwise
        """
        return not self.is_any_connected_channel_healthy_and_not_closed()

    def invoke_resource_creation(self) -> bool:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.

        :return: True if operation done, False if still in progress
        """

        self._resources_created = ensure_type(self._create_resources(), bool)
        return self._resources_created

    def invoke_resource_deletion(self) -> bool:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.

        :return: True if operation done, False if still in progress
        """

        self._resources_deleted = ensure_type(self._delete_resources(), bool)
        return self._resources_deleted

    def invoke_open_channel(self) -> bool:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.

        :return: True if operation done, False if still in progress
        """

        if not ensure_type(self._open_channel(), bool):
            return False

        self.invoke_sync_send_status_message(ChannelStatus.Opened)
        self.invoke_sync_status_update()

        if (not self._executor_started) and (not self.__silent_mode):
            if self._executor is None:
                self._executor = concurrent.futures.ThreadPoolExecutor(thread_name_prefix=self.get_unique_name())

            self._executor.submit(self.__executor_thread_handler)
            self._executor_started = True

        self._channel_opened = True

        return True

    def invoke_close_channel(self) -> bool:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.

        :return: True if operation done, False if still in progress
        """

        if not self._channel_opened:
            return True

        if self._executor_started:
            self._stopping_executor = True

            self._logger.debug("Stopping executor thread ...")
            while not self._executor_stopped.get():
                pass
            self._logger.debug("Executor thread stopped")

            if self._executor is not None:
                self._executor.shutdown()
                self._executor = None

        if self._channel_started and not self._channel_stopped:
            try:
                self.stop_channel()
            except Exception:
                self._logger.error("Exception at channel stopping, status message will not be sent.")
                self._logger.error(traceback.format_exc())
                self.stop_channel(False)

        try:
            self.invoke_sync_send_status_message(ChannelStatus.Closed)
        except Exception:
            self._logger.error("Exception at channel closing, status message will not be sent.")
            self._logger.error(traceback.format_exc())

        try:
            self.invoke_sync_status_update()
        except Exception:
            self._logger.error("Exception at status update in closing.")
            self._logger.error(traceback.format_exc())

        # If implemented logic is unfinished then channel is still open
        self._channel_opened = not ensure_type(self._close_channel(), bool)

        # If channel is still open then the closing procedure is unfinished
        return not self._channel_opened

    def invoke_configure_channel(self, channel_configuration: dict) -> None:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.
        """

        self._configuration.update(channel_configuration)

        self._configure_channel(channel_configuration)

        if ChannelBase.ParamKeyMetricsEnabled in self._configuration:
            self._metrics_enabled = self._configuration[ChannelBase.ParamKeyMetricsEnabled]

        if ChannelBase.ParamKeyLogLevel in self._configuration:
            self._log_level = self._configuration[ChannelBase.ParamKeyLogLevel]

        self._logger.set_log_level(logging.getLevelName(self._log_level))

    def invoke_sync_send_status_message(self, status: ChannelStatus, payload: Any = None) -> None:
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.
        """

        if not self.__silent_mode:
            self._send_status_message(str(ChannelStatusMessage(self._channel_name,
                                                               self._context.get_full_name(),
                                                               self._context.get_group_name(),
                                                               self._context.get_group_index(),
                                                               status, payload)))

    """
    This invoker is just for being consistent with the Java part, since in Java this method is synchronized due
    to the fact that it can be invoked from different threads. In python however this should not be an issue.
    """
    def invoke_sync_status_update(self):
        """
        An invoker method that encapsulates the actual implementation. This method
        MUST be called instead of the implemented method directly to ensure proper
        channel functionality.
        """

        string_status_messages = self._retrieve_status_messages()

        status_messages: list[ChannelStatusMessage] = list()

        if string_status_messages is not None:
            with self._channel_state_update_lock:
                for string_status_message in string_status_messages:
                    status_message = ChannelStatusMessage.create_from_json(string_status_message)
                    status_messages.append(status_message)

                    # Needs to be made sure that we don't maintain a monitor for ourselves
                    if (status_message.channel_name != self._channel_name) or \
                            (status_message.channel_context_name != self._context.get_full_name()):
                        if status_message.get_channel_unique_name() not in self._status_map:
                            self._status_map[status_message.get_channel_unique_name()] = \
                                ChannelStatusMonitor(status_message.channel_name,
                                                     status_message.channel_context_name,
                                                     status_message.channel_group_name,
                                                     self._logger)

                            self.on_new_channel_status_monitor(
                                self._status_map[status_message.get_channel_unique_name()])

                        self._status_map[status_message.get_channel_unique_name()].update(status_message)

            for callback in self._on_status_message_received_callbacks:
                try:
                    callback(status_messages)
                except Exception as e:
                    self._logger.error(f"Error at status message callback: {e}")
                    self._logger.error(f"{traceback.format_exc()}")

    def on_status_message_received(self, callback: Callable[[list[ChannelStatusMessage]], None]) -> None:
        """
        Adds a callback to the set of callbacks, which will be executed upon status
        message received.

        :param callback: callback Callable[[list[str]], None]
        """

        self._on_status_message_received_callbacks.add(callback)

    def on_new_channel_status_monitor(self, status_monitor: ChannelStatusMonitor) -> None:
        """
        This method can be overridden to handle the creation of the new ChannelStatusMonitor e.g.
        registering callbacks.

        :param status_monitor: ChannelStatusMonitor object newly created
        """
        pass

    def on_status_message_send(self):
        """
        This method can be implemented to hook into the event of sending status message.
        For example one can use it to calculate aggregated metrics between 2 sending and
        attach the metrics to the healthCheckPayload.
        VERY IMPORTANT NOTE: keep the runtime as low as possible, because the higher the
        runtime the more load on the status message sender thread, which might cause
        double activation.
        """
        pass

    # ======================= Meant to be private methods =======================

    """
    This method implements the logic to be executed in the background thread.
    """
    def __executor_thread_handler(self):
        self._logger.debug("Starting status sender thread ...")
        while not self._stopping_executor:

            control_loop_start_time = current_time_millis()

            try:
                self.on_status_message_send()

                self.invoke_sync_send_status_message(ChannelStatus.HealthCheck, self._health_check_payload)

                if self._metrics_enabled:
                    self._logger.debug(f"Metrics: {json.dumps(self._health_check_payload)}")

                self.invoke_sync_status_update()

                # Resetting the patience timer
                self._control_loop_exception_timer = 0
            except Exception:
                if 0 == self._control_loop_exception_timer:
                    self._control_loop_exception_timer = current_time_millis()

                # Note that we ignore all exceptions here, since we cannot risk to close this loop
                # upon a recoverable exception. Hence, we are only plotting the error.
                self._logger.error(f"Unhandled exception in control loop. Timeout in "
                                   f"{current_time_millis() - self._control_loop_exception_timer}/"
                                   f"{ChannelBase.ControlLoopExceptionTimeoutInMs}")
                self._logger.error(traceback.format_exc())

                if ChannelBase.ControlLoopExceptionTimeoutInMs < \
                        (current_time_millis() - self._control_loop_exception_timer):
                    self._stopping_executor = True
            finally:
                control_loop_duration = current_time_millis() - control_loop_start_time

                """
                Calculating the necessary sleep time for the loop. This is necessary in
                order not to block to long, since if the implementation takes 2 sec and
                the loop waits additional 2 secs, then it can miss important information.
                So, we consider the implementation's runtime e.g. if it took 1500ms then,
                the sleep will wait 2000-1500=500ms. Notice that it can never gol below 0.
                """
                if ChannelBase.DefaultStatusThreadIntervalInMs > control_loop_duration:
                    time.sleep((ChannelBase.DefaultStatusThreadIntervalInMs - control_loop_duration) / 1000.0)

        try:
            # Sending one last status message to make sure that all relevant payloads are transferred.
            self.invoke_sync_send_status_message(ChannelStatus.HealthCheck, self._health_check_payload)
        except Exception as ignored:
            self._logger.debug(str(ignored))

        self._executor_stopped.set(True)
        self._stopping_executor = False
        self._executor_started = False
        self._logger.debug("Status sender thread stopped.")



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import concurrent.futures
from typing import Any, TYPE_CHECKING

from pypz.core.channels.io import ChannelReader, ChannelWriter

if TYPE_CHECKING:
    from pypz.core.specs.plugin import InputPortPlugin, OutputPortPlugin


class BlankChannelReader(ChannelReader):
    def __init__(self, channel_name: str,
                 context: 'InputPortPlugin',
                 executor: concurrent.futures.ThreadPoolExecutor = None,
                 **kwargs):
        super().__init__(channel_name, context, executor, **kwargs)

    def _load_input_record_offset(self) -> int:
        return 0

    def has_records(self) -> bool:
        return False

    def _read_records(self) -> list[Any]:
        return []

    def _commit_offset(self, offset: int) -> None:
        pass

    def _create_resources(self) -> bool:
        return True

    def _delete_resources(self) -> bool:
        return True

    def _open_channel(self) -> bool:
        return True

    def _close_channel(self) -> bool:
        return True

    def _configure_channel(self, channel_configuration: dict) -> None:
        pass

    def _send_status_message(self, message: str) -> None:
        pass

    def _retrieve_status_messages(self) -> list:
        return []


class BlankChannelWriter(ChannelWriter):

    def __init__(self, channel_name: str,
                 context: 'OutputPortPlugin',
                 executor: concurrent.futures.ThreadPoolExecutor = None,
                 **kwargs):
        super().__init__(channel_name, context, executor, **kwargs)

    def _write_records(self, records: list[Any]) -> None:
        pass

    def _create_resources(self) -> bool:
        return True

    def _delete_resources(self) -> bool:
        return True

    def _open_channel(self) -> bool:
        return True

    def _close_channel(self) -> bool:
        return True

    def _configure_channel(self, channel_configuration: dict) -> None:
        pass

    def _send_status_message(self, message: str) -> None:
        pass

    def _retrieve_status_messages(self) -> list:
        return []




# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import Any, Optional

from pypz.abstracts.channel_ports import ChannelInputPort, ChannelOutputPort
from pypz.core.channels.misc import BlankChannelReader, BlankChannelWriter


class BlankChannelInputPort(ChannelInputPort):

    def __init__(self, name: str = None, schema: Any = None, group_mode: bool = False, *args, **kwargs):
        super().__init__(name, schema, group_mode, BlankChannelReader, *args, **kwargs)


class BlankChannelOutputPort(ChannelOutputPort):
    def __init__(self, name: str = None, schema: Optional[Any] = None, *args, **kwargs):
        super().__init__(name, schema, BlankChannelWriter, *args, **kwargs)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import logging
import sys
from abc import ABC, abstractmethod
from typing import Any, Optional

import colorlog
from colorlog import ColoredFormatter


class ContextLoggerInterface(ABC):
    """
    This interface provides the necessary methods to be able to implement
    the context logging functionality. Context logging means that there
    is exactly one logger, which will be (re)used in the given context.
    """

    @abstractmethod
    def _error(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        pass

    @abstractmethod
    def _warn(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        pass

    @abstractmethod
    def _info(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        pass

    @abstractmethod
    def _debug(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        pass

    def set_log_level(self, log_level: Any) -> None:
        pass


class ContextLogger(ContextLoggerInterface):
    """
    This class has the responsibility to provide a simple logging interface, which
    hides the complexity of maintaining the context information. The user can
    call the familiar methods without knowing, how the context information are
    handled by the protected methods.

    By expecting a logger in the constructor we are allowing to reuse existing loggers
    from higher contexts i.e., a top-level instance can provide it's logger to the
    nested instances along with the context information. The nested instance can
    further forward to it's nested instances and so on. Each instance provides itself
    to the context stack so, if a logger method of the actual context will be called,
    then the logger itself will get the context information automatically.

    :param logger: a logger implementation that will be used to route the messages to
    :param context_stack: the information about the current and previous contexts
    """

    def __init__(self, logger: ContextLoggerInterface, *context_stack: str):
        self._logger: ContextLoggerInterface = logger
        """
        The provided logger to use on this context
        """

        self._context_stack: list[str] = [*context_stack]
        """
        The context stack, which contains the actual and all parent contexts
        """

    def get_context_stack(self) -> list[str]:
        return self._context_stack

    def _error(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger._error(event, context_stack, *args, **kw)

    def _warn(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger._warn(event, context_stack, *args, **kw)

    def _info(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger._info(event, context_stack, *args, **kw)

    def _debug(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger._debug(event, context_stack, *args, **kw)

    def set_log_level(self, log_level: str | int) -> None:
        self._logger.set_log_level(log_level)

    def error(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        self._error(event, self._context_stack, *args, **kw)

    def warn(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        self._warn(event, self._context_stack, *args, **kw)

    def info(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        self._info(event, self._context_stack, *args, **kw)

    def debug(self, event: Optional[str] = None, *args: Any, **kw: Any) -> Any:
        self._debug(event, self._context_stack, *args, **kw)


class DefaultContextLogger(ContextLoggerInterface):
    """
    The default implementation of the :class:`ContextLoggerInterface <pypz.core.commons.loggers.ContextLoggerInterface>`,
    which will send log messages to the standard out.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None):
        logging.basicConfig()
        self._logger: logging.Logger = colorlog.getLogger(name)
        self._logger.propagate = False

        if not self._logger.handlers:
            handler = colorlog.StreamHandler(sys.stdout)
            formatter = ColoredFormatter(
                "%(log_color)s%(levelname)s | %(asctime)s | %(name)s | %(context)s | %(message)s",
                log_colors={
                    'DEBUG': 'cyan',
                    'INFO': 'light_black',
                    'WARNING': 'yellow',
                    'ERROR': 'red',
                    'CRITICAL': 'red,bg_white',
                }
            )
            handler.setFormatter(formatter)
            self._logger.addHandler(handler)

    def _error(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger.error(event,
                           extra={
                               "context": " | ".join(context_stack) if context_stack is not None else None
                           },
                           *args, **kw)

    def _warn(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger.warning(event,
                             extra={
                                 "context": " | ".join(context_stack) if context_stack is not None else None
                             },
                             *args, **kw)

    def _info(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger.info(event,
                          extra={
                              "context": " | ".join(context_stack) if context_stack is not None else None
                          },
                          *args, **kw)

    def _debug(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
        self._logger.debug(event,
                           extra={
                               "context": " | ".join(context_stack) if context_stack is not None else None
                           },
                           *args, **kw)

    def set_log_level(self, log_level: str | int) -> None:
        if isinstance(log_level, str):
            self._logger.setLevel(logging.getLevelName(log_level.upper()))
        elif isinstance(log_level, int):
            self._logger.setLevel(log_level)
        else:
            raise TypeError(f"Invalid log_level type: {log_level}")
